/*************************************************************************
 *			       						 *
 * md5c_opt.c					         		 *
 * June, 1995							         *
 *									 *
 * Copyright (c) 1995 by the University of Southern California		 *
 * All rights reserved.							 *
 *									 *
 * Permission to use, copy, modify, and distribute this software and its *
 * documentation in source and binary forms for non-commercial purposes	 *
 * and without fee is hereby granted, provided that the above copyright	 *
 * notice appear in all copies and that both the copyright notice and	 *
 * this permission notice appear in supporting documentation, and that	 *
 * any documentation, advertising materials, and other materials related *
 * to such distribution and use acknowledge that the software was	 *
 * developed by the University of Southern California, Information	 *
 * Sciences Institute.  The name of the University may not be used to	 *
 * endorse or promote products derived from this software without	 *
 * specific prior written permission.					 *
 *									 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about	 *
 * the suitability of this software for any purpose.  THIS SOFTWARE IS	 *
 * PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,	 *
 * INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF		 *
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.		 *
 *									 *
 * Other copyrights might apply to parts of this software and are so	 *
 * noted when applicable.						 *
 *									 *
 *************************************************************************/

/* 
   This is a _PRELIMINARY_ version of the "Alternate Hash Algorithm,"
   described in the appendix of:
   	"Performance Analysis of MD5",
	J. Touch,
	to appear in Proc. Sigcomm, August, 1995.
	For further information see: 
		ftp://ftp.isi.edu/pub/hpcc-papers/touch/sigcomm95.ps.Z
		http://www.isi.edu/div7/atomic2/encryptor.html

   See the README-AHA for more information

   This algorithm was developed with Amir Herzberg and members of his
   research group at IBM T.J. Watson Labs, Hawthorne, NY.

   This code is distributed as a replacement for the md5c_opt.c code,
   part of the optimized MD5 code available at:
   	ftp://ftp.isi.edu/pub/hpcc-papers/touch/md5-opt.tar

   It is distributed as:
   	ftp://ftp.isi.edu/pub/hpcc-papers/touch/aha-v0.tar
   
*/

/* MD5C.C - RSA Data Security, Inc., MD5 message-digest algorithm
 */

/* Copyright (C) 1991-2, RSA Data Security, Inc. Created 1991. All
rights reserved.

License to copy and use this software is granted provided that it
is identified as the "RSA Data Security, Inc. MD5 Message-Digest
Algorithm" in all material mentioning or referencing this software
or this function.

License is also granted to make and use derivative works provided
that such works are identified as "derived from the RSA Data
Security, Inc. MD5 Message-Digest Algorithm" in all material
mentioning or referencing the derived work.

RSA Data Security, Inc. makes no representations concerning either
the merchantability of this software or the suitability of this
software for any particular purpose. It is provided "as is"
without express or implied warranty of any kind.

These notices must be retained in any copies of any part of this
documentation and/or software.
 */

#include "global.h"
#include "md5.h"

#include <stdio.h>

#include <memory.h>

static void MD5Transform PROTO_LIST ((UINT4 [4], unsigned char [64]));
static void Encode PROTO_LIST
  ((unsigned char *, UINT4 *, unsigned int));
static void Decode PROTO_LIST
  ((UINT4 *, unsigned char *, unsigned int));
static void MD5_memcpy PROTO_LIST ((POINTER, POINTER, unsigned int));
static void MD5_memset PROTO_LIST ((POINTER, int, unsigned int));

static unsigned char PADDING[64] = {
  0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

/* ROTATE_LEFT rotates x left n bits.
 */
#define ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32-(n))))

/* FF, GG, HH, and II transformations for rounds 1, 2, 3, and 4.
Rotation is separate from addition to prevent recomputation.
 */

#define CHARSIZE 256
UINT4 l_arr[CHARSIZE];

#if (defined(NATIVE) || ((!defined(i386)) && (!defined(__alpha))))
#define swapbyte(src,dst) { }
#elif ((!defined(hpux)) & (!defined(mc68000)))
#define swapbyte(src,dst) { \
     register UINT4 tmp; \
     dst = ROTATE_LEFT((src),16); \
     tmp = (dst) >> 8; \
     tmp &= 0x00ff00ff; \
     dst &= 0x00ff00ff; \
     dst <<= 8; \
     (dst) = dst | tmp; \
     }
#else
#define swapbyte(src,dst) { \
     register UINT4 out,other;
     (dst) = (ROTATE_LEFT((src),8) & 0x00ff00ff) | ROTATE_LEFT((src) & 0x00ff00ff,24); \
     }
#endif

#if (defined(i386) || defined(__alpha))

#define HASHWORD(a, b, c, d, x) { \
 register UINT4 tmp; \
 tmp = (a) ^ (x); \
 (a) ^= l_arr[(tmp >> 24)]; \
 (b) ^= l_arr[(tmp >> 16) & 0xff]; \
 (c) ^= l_arr[(tmp >>  8) & 0xff]; \
 (d) ^= l_arr[tmp & 0xff]; \
  }

#else

#define HASHWORD(a, b, c, d, x) { \
 register UINT4 tmp; \
 tmp = (a) ^ (x); \
 (a) ^= l_arr[tmp & 0xff]; \
 (b) ^= l_arr[(tmp >>  8) & 0xff]; \
 (c) ^= l_arr[(tmp >> 16) & 0xff]; \
 (d) ^= l_arr[(tmp >> 24)]; \
  }

#endif

/* MD5 initialization. Begins an MD5 operation, writing a new context.
 */
void MD5Init (context)
MD5_CTX *context;                                        /* context */
{
  context->count[0] = context->count[1] = 0;
  /* Load magic initialization constants.
*/
  context->state[0] = 0x67452301;
  context->state[1] = 0xefcdab89;
  context->state[2] = 0x98badcfe;
  context->state[3] = 0x10325476;

/* initialize the lookup array - use a seed based on the
   session key here to make the hash session-dependent 

   the cost of this table generation can be reduced by 
   updating only part of the table, or XORing the table
   with the session key

   something should be done here to make sure that the table
   is "overloaded", i.e., non-reversible - this can be done
   by copying entries 

   Because the array is indexed byte-wise, there's no need
   to swap entries (as there would be for half-word indexing)

*/

  {
    long i;

    srand(1); /* initialize the pseudo-random generator */
    for (i=0; i<CHARSIZE; i++) {
/*      l_arr[i] = rand(); */
/* fix table values for debugging */
      l_arr[i] = (i * 0x01010101) ^ (0xa55aaa55 << (i & 0x01));
      swapbyte(l_arr[i],l_arr[i]);
    }
#ifdef debug
    if (print_debug) {
      printf("\n\nL_ARR TABLE\n");
      for (i=0; i< CHARSIZE; i+=4) {
	printf("%3d: %8x %8x %8x %8x\n",i,l_arr[i],l_arr[i+1],l_arr[i+2],l_arr[i+3]);
      }
    }
#endif
  }
}

/* MD5 block update operation. Continues an MD5 message-digest
  operation, processing another message block, and updating the
  context.
 */
void MD5Update (context, input, inputLen)
MD5_CTX *context;                                        /* context */
unsigned char *input;                                /* input block */
unsigned int inputLen;                     /* length of input block */
{
  unsigned int i, index, partLen;

  /* Compute number of bytes mod 64 */
  index = (unsigned int)((context->count[0] >> 3) & 0x3F);

  /* Update number of bits */
  if ((context->count[0] += ((UINT4)inputLen << 3))
   < ((UINT4)inputLen << 3))
 context->count[1]++;
  context->count[1] += ((UINT4)inputLen >> 29);

  partLen = 64 - index;

  /* Transform as many times as possible.
*/
  if (inputLen >= partLen) {
 MD5_memcpy
   ((POINTER)&context->buffer[index], (POINTER)input, partLen);
 MD5Transform (context->state, context->buffer);

 for (i = partLen; i + 63 < inputLen; i += 64)
   MD5Transform (context->state, &input[i]);

 index = 0;
  }
  else
 i = 0;

  /* Buffer remaining input */
  MD5_memcpy
 ((POINTER)&context->buffer[index], (POINTER)&input[i],
  inputLen-i);
}

/* MD5 finalization. Ends an MD5 message-digest operation, writing the
  the message digest and zeroizing the context.
 */
void MD5Final (digest, context)
unsigned char digest[16];                         /* message digest */
MD5_CTX *context;                                       /* context */
{
  unsigned char bits[8];
  unsigned int index, padLen;

  /* Save number of bits */
  Encode (bits, context->count, 8);

  /* Pad out to 56 mod 64.
*/
  index = (unsigned int)((context->count[0] >> 3) & 0x3f);
  padLen = (index < 56) ? (56 - index) : (120 - index);
  MD5Update (context, PADDING, padLen);

  /* Append length (before padding) */
  MD5Update (context, bits, 8);
  /* Store state in digest */
  Encode (digest, context->state, 16);

  /* Zeroize sensitive information.
*/
  MD5_memset ((POINTER)context, 0, sizeof (*context));
}

/* MD5 basic transformation. Transforms state based on block.
 */
static void MD5Transform (state, block)
UINT4 state[4];
unsigned char block[64];
{
  register UINT4 a = state[0], b = state[1], c = state[2], d = state[3];

  swapbyte(a,a);
  swapbyte(b,b);
  swapbyte(c,c);
  swapbyte(d,d);

#ifdef debug
if (print_debug) printf("\n\nPRE-TRANSFORM STATE\nA:%8lx B:%8lx C:%8lx D:%8lx\n",a,b,c,d);
#endif

  HASHWORD(a,b,c,d,((UINT4 *)block)[ 0]);
  HASHWORD(b,a,d,c,((UINT4 *)block)[ 1]);
  HASHWORD(c,d,a,b,((UINT4 *)block)[ 2]);
  HASHWORD(d,c,b,a,((UINT4 *)block)[ 3]);

  HASHWORD(a,c,b,d,((UINT4 *)block)[ 4]);
  HASHWORD(c,a,d,b,((UINT4 *)block)[ 5]);
  HASHWORD(b,d,a,c,((UINT4 *)block)[ 6]);
  HASHWORD(d,b,c,a,((UINT4 *)block)[ 7]);

  HASHWORD(a,b,c,d,((UINT4 *)block)[ 8]);
  HASHWORD(b,a,d,c,((UINT4 *)block)[ 9]);
  HASHWORD(c,d,a,b,((UINT4 *)block)[10]);
  HASHWORD(d,c,b,a,((UINT4 *)block)[11]);

  HASHWORD(a,d,b,c,((UINT4 *)block)[12]);
  HASHWORD(d,a,c,b,((UINT4 *)block)[13]);
  HASHWORD(b,c,a,d,((UINT4 *)block)[14]);
  HASHWORD(c,b,d,a,((UINT4 *)block)[15]);

/* make the state in the standard order */

#ifdef debug
  if (print_debug) {
    printf("s[0] = %8x \ta = %8x\n",state[0],a);
    printf("s[1] = %8x \tb = %8x\n",state[1],b);
    printf("s[2] = %8x \tc = %8x\n",state[2],c);
    printf("s[3] = %8x \td = %8x\n",state[3],d);
    
  }
#endif

  swapbyte(a,a);
  swapbyte(b,b);
  swapbyte(c,c);
  swapbyte(d,d);

  state[0] += a;
  state[1] += b;
  state[2] += c;
  state[3] += d;

#ifdef debug
  if (print_debug) {
    printf("\n\nPOST-TRANSFORM STATE\n    %8lx %8lx %8lx %8lx\n",a,b,c,d);
    printf("\n\nTOTAL STATE\n\t%8lx %8lx %8lx %8lx\n",state[0],state[1],state[2],state[3]);
  }
#endif

}

/* Encodes input (UINT4) into output (unsigned char). Assumes len is
  a multiple of 4.
 */
static void Encode (output, input, len)
unsigned char *output;
UINT4 *input;
unsigned int len;
{
  unsigned int i, j;

  for (i = 0, j = 0; j < len; i++, j += 4) {
 output[j] = (unsigned char)(input[i] & 0xff);
 output[j+1] = (unsigned char)((input[i] >> 8) & 0xff);
 output[j+2] = (unsigned char)((input[i] >> 16) & 0xff);
 output[j+3] = (unsigned char)((input[i] >> 24) & 0xff);
  }
}

/* Decodes input (unsigned char) into output (UINT4). Assumes len is
  a multiple of 4.
 */
static void Decode (output, input, len)
UINT4 *output;
unsigned char *input;
unsigned int len;
{
  unsigned int i, j;

/*
  for (i = 0, j = 0; j < len; i++, j += 4)
 output[i] = ((UINT4)input[j]) | (((UINT4)input[j+1]) << 8) |
   (((UINT4)input[j+2]) << 16) | (((UINT4)input[j+3]) << 24);
*/

swapbyte(((UINT4 *)input)[0],output[0]);
swapbyte(((UINT4 *)input)[1],output[1]);
swapbyte(((UINT4 *)input)[2],output[2]);
swapbyte(((UINT4 *)input)[3],output[3]);
swapbyte(((UINT4 *)input)[4],output[4]);
swapbyte(((UINT4 *)input)[5],output[5]);
swapbyte(((UINT4 *)input)[6],output[6]);
swapbyte(((UINT4 *)input)[7],output[7]);
swapbyte(((UINT4 *)input)[8],output[8]);
swapbyte(((UINT4 *)input)[9],output[9]);
swapbyte(((UINT4 *)input)[10],output[10]);
swapbyte(((UINT4 *)input)[11],output[11]);
swapbyte(((UINT4 *)input)[12],output[12]);
swapbyte(((UINT4 *)input)[13],output[13]);
swapbyte(((UINT4 *)input)[14],output[14]);
swapbyte(((UINT4 *)input)[15],output[15]);

}

/* Note: Replace "for loop" with standard memcpy if possible.
 */

static void MD5_memcpy (output, input, len)
POINTER output;
POINTER input;
unsigned int len;
{
/*
  unsigned int i;

  for (i = 0; i < len; i++)
 output[i] = input[i];
*/
  memcpy(output,input,len);
}

/* Note: Replace "for loop" with standard memset if possible.
 */
static void MD5_memset (output, value, len)
POINTER output;
int value;
unsigned int len;
{
/*
  unsigned int i;

  for (i = 0; i < len; i++)
 ((char *)output)[i] = (char)value;
*/
  memset(output,value,len);
}
